/*
 * Copyright (C) 2009--2012,2014,2015, 2016  Kipp Cannon <kipp.cannon@ligo.org>, Madeline Wade <madeline.wade@ligo.org>, Aaron Viets <aaron.viets@ligo.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */


/**
 * SECTION:gstlal_smoothkappas
 * @short_description:  Smooths the calibration factors (kappas) using a 
 * running median.
 *
 * This element smooths the kappas using a running median of an array 
 * whose size is set by the property array-size. When a new raw value
 * is entered into the array, it replaces the oldest value in the array
 * (first in, first out). When this element receives a gap as input, it 
 * will output a default kappa value (set by the property default-kappa)
 * until it receives a buffer that is not flagged as a gap.
 */


/*
 * ============================================================================
 *
 *				  Preamble
 *
 * ============================================================================
 */


/*
 * stuff from C
 */


#include <math.h>
#include <string.h>
#include <complex.h>


/*
 *  stuff from gobject/gstreamer
 */


#include <glib.h>
#include <gst/gst.h>
#include <gst/audio/audio.h>
#include <gst/base/gstbasetransform.h>


/*
 * our own stuff
 */


#include <gstlal/gstlal.h>
#include <gstlal/gstlal_debug.h>
#include <gstlal/gstlal_audio_info.h>
#include <gstlal_smoothkappas.h>


/*
 * ============================================================================
 *
 *				 Parameters
 *
 * ============================================================================
 */


/*
 * ============================================================================
 *
 *			   GStreamer Boiler Plate
 *
 * ============================================================================
 */


static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
	GST_BASE_TRANSFORM_SINK_NAME,
	GST_PAD_SINK,
	GST_PAD_ALWAYS,
	GST_STATIC_CAPS(
		"audio/x-raw, " \
		"rate = (int) [1, MAX], " \
		"channels = (int) 1, " \
		"format = (string) { " GST_AUDIO_NE(F32) ", " GST_AUDIO_NE(F64) ", " GST_AUDIO_NE(Z64) ", " GST_AUDIO_NE(Z128) " }, " \
		"layout = (string) interleaved, " \
		"channel-mask = (bitmask) 0"
	)
);


static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
	GST_BASE_TRANSFORM_SRC_NAME,
	GST_PAD_SRC,
	GST_PAD_ALWAYS,
	GST_STATIC_CAPS(
		"audio/x-raw, " \
		"rate = (int) [1, MAX], " \
		"channels = (int) 1, " \
		"format = (string) { " GST_AUDIO_NE(F32) ", " GST_AUDIO_NE(F64) ", " GST_AUDIO_NE(Z64) ", " GST_AUDIO_NE(Z128) " }, " \
		"layout = (string) interleaved, " \
		"channel-mask = (bitmask) 0"
	)
);


#define GST_CAT_DEFAULT gstlal_smoothkappas_debug
GST_DEBUG_CATEGORY_STATIC(GST_CAT_DEFAULT);


G_DEFINE_TYPE_WITH_CODE(
	GSTLALSmoothKappas,
	gstlal_smoothkappas,
	GST_TYPE_BASE_TRANSFORM,
	GST_DEBUG_CATEGORY_INIT(GST_CAT_DEFAULT, "lal_smoothkappas", 0, "lal_smoothkappas element")
);


/*
 * ============================================================================
 *
 *				 Utilities
 *
 * ============================================================================
 */


/*
 * set the metadata on an output buffer
 */


static void set_metadata(GSTLALSmoothKappas *element, GstBuffer *buf, guint64 outsamples)
{
	GST_BUFFER_OFFSET(buf) = element->next_out_offset;
	element->next_out_offset += outsamples;
	GST_BUFFER_OFFSET_END(buf) = element->next_out_offset;
	GST_BUFFER_PTS(buf) = element->t0 + gst_util_uint64_scale_int_round(GST_BUFFER_OFFSET(buf) - element->offset0, GST_SECOND, element->rate);
	GST_BUFFER_DURATION(buf) = element->t0 + gst_util_uint64_scale_int_round(GST_BUFFER_OFFSET_END(buf) - element->offset0, GST_SECOND, element->rate) - GST_BUFFER_PTS(buf);
	GST_BUFFER_FLAG_UNSET(buf, GST_BUFFER_FLAG_GAP);
	if(G_UNLIKELY(element->need_discont)) {
		GST_BUFFER_FLAG_SET(buf, GST_BUFFER_FLAG_DISCONT);
		element->need_discont = FALSE;
	}
}


static void get_new_median(double new_element, double *fifo_array, double *current_median, gint array_size, int *index_re, int *index_im, gboolean array_is_imaginary, int samples_in_filter, gboolean no_default, int future_unused) {
	if(array_is_imaginary) {
		fifo_array[*index_im] = new_element;
		(*index_im)++;
		*index_im %= array_size;
	} else {
		fifo_array[*index_re] = new_element;
		(*index_re)++;
		*index_re %= array_size;
	}

	int j, jj, j_start, num_samples, number_less, number_greater, number_equal;
	double greater, less, Greater, Less;
	number_less = 0;
	number_equal = 0;
	number_greater = 0;
	greater = Greater = G_MAXDOUBLE;
	less = Less = -G_MAXDOUBLE;
	num_samples = (no_default && samples_in_filter < array_size) ? samples_in_filter : array_size;
	j_start = (array_is_imaginary ? *index_im : *index_re) + array_size - num_samples - future_unused;
	for(j = j_start; j < j_start + num_samples; j++) {
		jj = (j + array_size) % array_size;
		if(fifo_array[jj] < *current_median) {
			number_less++;
			if(fifo_array[jj] >= less) {
				Less = less;
				less = fifo_array[jj];
			} else if(fifo_array[jj] > Less)
				Less = fifo_array[jj];
		} else if(fifo_array[jj] == *current_median)
			number_equal++;
		else if(fifo_array[jj] > *current_median) {
			number_greater++;
			if(fifo_array[jj] <= greater) {
				Greater = greater;
				greater = fifo_array[jj];
			} else if(fifo_array[jj] < Greater)
				Greater = fifo_array[jj];
		} else
			g_assert_not_reached();
	}
	g_assert_cmpint(number_less + number_equal + number_greater, ==, num_samples);

	if(num_samples % 2) {
		if(number_greater > 1 + num_samples / 2)
			*current_median = Greater;
		else if(number_greater > num_samples / 2)
			*current_median = greater;
		else if(number_less > 1 + num_samples / 2)
			*current_median = Less;
		else if(number_less > num_samples / 2)
			*current_median = less;
	} else {
		if(number_greater > num_samples / 2)
			*current_median = (Greater + greater) / 2;
		else if(number_less > num_samples / 2)
			*current_median = (Less + less) / 2;
		else if(number_greater == num_samples / 2 && number_less == num_samples / 2)
			*current_median = (greater + less) / 2;
		else if(number_greater == num_samples / 2)
			*current_median = (greater + *current_median) / 2;
		else if(number_less == num_samples / 2)
			*current_median = (less + *current_median) / 2;
	}

	return;
}


static double get_average(double new_element, double *fifo_array, gint array_size, int *index_re, int *index_im, gboolean array_is_imaginary, int samples_in_filter, gboolean no_default) {
	if(array_is_imaginary) {
		fifo_array[*index_im] = new_element;
		(*index_im)++;
		*index_im %= array_size;
	} else {
		fifo_array[*index_re] = new_element;
		(*index_re)++;
		*index_re %= array_size;
	}
	double sum = 0;
	int i, ii, i_start, num_samples;
	num_samples = (no_default && samples_in_filter < array_size) ? samples_in_filter : array_size;
	i_start = (array_is_imaginary ? *index_im : *index_re) + array_size - num_samples;
	for(i = i_start; i < i_start + num_samples; i++) {
		ii = i % array_size;
		sum += fifo_array[ii];
	}
	return sum / num_samples;
}


/*
 * Reset the running median and running average calculation.  This deletes data stored
 * in the median and average arrays to start the calculation over.  The property
 * min-array-size determines the number of samples to be kept in the median and
 * average arrays.
 */


static void reset(GSTLALSmoothKappas *element) {
	/* Calculate number of samples currently in median array that are too new to be used yet.  We want the truncated array to contain samples with timestamps closest to that of the next output sample */
	if(element->array_size * element->filter_latency < element->min_array_size / 2)
		element->future_unused = 0;
	else if(element->array_size * element->filter_latency > element->samples_in_filter - element->min_array_size / 2)
		element->future_unused = element->samples_in_filter - element->min_array_size;
	else
		element->future_unused = (int) (element->array_size * element->filter_latency - element->min_array_size / 2);
	g_assert_cmpint(element->future_unused, >=, 0);
	/* Drop samples to make truncated array */
	element->samples_in_filter = element->min_array_size;
	int i, num_default, ii, i_start, i_stop, num_less, num_greater, num_equal, num_samples;
	double greater, less, replace_value_re, replace_value_im;
	num_samples = (element->no_default && element->samples_in_filter < element->array_size) ? element->samples_in_filter : element->array_size;

	/* Update median array and bookkeeping */
	if(element->array_size > element->samples_in_filter) {
		num_default = element->array_size - element->min_array_size;
		/* Replace removed values with default values */
		if(!element->default_to_median && !element->no_default) {
			for(i = element->index_re - element->future_unused + element->array_size; i < element->index_re - element->future_unused + element->array_size + num_default; i++)
				element->fifo_array_re[i % element->array_size] = element->default_kappa_re;
			for(i = element->index_im - element->future_unused + element->array_size; i < element->index_im - element->future_unused + element->array_size + num_default; i++)
				element->fifo_array_im[i % element->array_size] = element->default_kappa_im;
		}
		/* Find new median */
		num_less = 0;
		num_equal = 0;
		num_greater = 0;
		greater = G_MAXDOUBLE;
		less = -G_MAXDOUBLE;
		/* Real part */
		while(num_less + num_equal + num_greater == 0 || num_greater > num_samples / 2 || num_less > num_samples / 2) {
			num_less = 0;
			num_equal = 0;
			num_greater = 0;
			greater = G_MAXDOUBLE;
			less = -G_MAXDOUBLE;
			i_start = (element->index_re + 2 * element->array_size - element->samples_in_filter - element->future_unused) % element->array_size;
			i_stop = i_start + (element->no_default ? element->samples_in_filter : element->array_size);
			for(i = i_start; i < i_stop; i++) {
				ii = i % element->array_size;
				if(element->fifo_array_re[ii] < element->current_median_re) {
					num_less++;
					if(element->fifo_array_re[ii] > less)
						less = element->fifo_array_re[ii];
				} else if(element->fifo_array_re[ii] == element->current_median_re)
					num_equal++;
				else {
					num_greater++;
					if(element->fifo_array_re[ii] < greater)
						greater = element->fifo_array_re[ii];
				}
			}
			if(num_less == num_samples / 2 && num_greater == num_samples / 2 && !(num_samples % 2))
				element->current_median_re = (less + greater) / 2;
			else if(num_less == num_samples / 2 && !(num_samples % 2))
				element->current_median_re = (less + element->current_median_re) / 2;
			else if(num_greater == num_samples / 2 && !(num_samples % 2))
				element->current_median_re = (element->current_median_re + greater) / 2;
			else if(num_greater > num_samples / 2)
				element->current_median_re = greater;
			else if(num_less > num_samples / 2)
				element->current_median_re = less;
			/* In all other cases, do nothing */
		}
		num_less = 0;
		num_equal = 0;
		num_greater = 0;
		greater = G_MAXDOUBLE;
		less = -G_MAXDOUBLE;
		/* Imaginary part */
		while(num_less + num_equal + num_greater == 0 || num_greater > num_samples / 2 || num_less > num_samples / 2) {
			num_less = 0;
			num_equal = 0;
			num_greater = 0;
			greater = G_MAXDOUBLE;
			less = -G_MAXDOUBLE;
			i_start = (element->index_im + 2 * element->array_size - element->samples_in_filter - element->future_unused) % element->array_size;
			i_stop = i_start + (element->no_default ? element->samples_in_filter : element->array_size);
			for(i = i_start; i < i_stop; i++) {
				ii = i % element->array_size;
				if(element->fifo_array_im[ii] < element->current_median_im) {
					num_less++;
					if(element->fifo_array_im[ii] > less)
						less = element->fifo_array_im[ii];
				} else if(element->fifo_array_im[ii] == element->current_median_im)
					num_equal++;
				else {
					num_greater++;
					if(element->fifo_array_im[ii] < greater)
						greater = element->fifo_array_im[ii];
				}
			}
			if(num_less == num_samples / 2 && num_greater == num_samples / 2 && !(num_samples % 2))
				element->current_median_im = (less + greater) / 2;
			else if(num_less == num_samples / 2 && !(num_samples % 2))
				element->current_median_im = (less + element->current_median_im) / 2;
			else if(num_greater == num_samples / 2 && !(num_samples % 2))
				element->current_median_im = (element->current_median_im + greater) / 2;
			else if(num_greater > num_samples / 2)
				element->current_median_im = greater;
			else if(num_less > num_samples / 2)
				element->current_median_im = less;
			/* In all other cases, do nothing */
		}
		if(element->default_to_median && !element->no_default) {
			/* Fill removed values with current median */
			for(i = element->index_re - element->future_unused + element->array_size; i < element->index_re - element->future_unused + element->array_size + num_default; i++)
				element->fifo_array_re[i % element->array_size] = element->current_median_re;
			for(i = element->index_im - element->future_unused + element->array_size; i < element->index_im - element->future_unused + element->array_size + num_default; i++)
				element->fifo_array_im[i % element->array_size] = element->current_median_im;
		}
	}
	if(element->future_unused > 0)
		(element->future_unused)--;

	/* Update average array and bookkeeping */
	if(element->avg_array_size > element->samples_in_filter) {
		num_default = element->avg_array_size - element->min_array_size;
		replace_value_re = element->default_to_median ? element->current_median_re : element->default_kappa_re;
		replace_value_im = element->default_to_median ? element->current_median_im : element->default_kappa_im;
		for(i = element->avg_index_re; i < element->avg_index_re + num_default; i++)
			element->avg_array_re[i % element->avg_array_size] = replace_value_re;
		for(i = element->avg_index_im; i < element->avg_index_im + num_default; i++)
			element->avg_array_im[i % element->avg_array_size] = replace_value_im;
	}

	return;
}


#define DEFINE_SMOOTH_BUFFER(DTYPE) \
static GstFlowReturn smooth_buffer_ ## DTYPE(const DTYPE *src, guint64 src_size, DTYPE *dst, guint64 dst_size, double *fifo_array, double *avg_array, double default_kappa, double *current_median, double maximum_offset, gint array_size, gint avg_array_size, int *index_re, int *index_im, int *avg_index_re, int *avg_index_im, int *num_bad_in_avg_re, gboolean gap, gboolean default_to_median, gboolean track_bad_kappa, int *samples_in_filter, gboolean no_default, guint64 *total_samples, gboolean reject_zeros, int *future_unused) { \
	guint64 i; \
	double new_element; \
	DTYPE new_avg; \
	for(i = 0; i < src_size; i++) { \
		if(gap || (double) *src > default_kappa + maximum_offset || (double) *src < default_kappa - maximum_offset || isnan(*src) || isinf(*src) || (reject_zeros && (double) *src == 0.0)) { \
			if(default_to_median) \
				new_element = *current_median; \
			else \
			new_element = default_kappa; \
 \
			if(*num_bad_in_avg_re <= avg_array_size) \
				(*num_bad_in_avg_re)++; \
		} else { \
			new_element = (double) *src; \
			*num_bad_in_avg_re = 0; \
		} \
 \
		/* Compute new median */ \
		get_new_median(new_element, fifo_array, current_median, array_size, index_re, index_im, FALSE, *samples_in_filter + (int) i + 1, no_default, *future_unused); \
		if(*future_unused > 0) \
			(*future_unused)--; \
		/* Compute new average */ \
		new_avg = (DTYPE) get_average(*current_median, avg_array, avg_array_size, avg_index_re, avg_index_im, FALSE, *samples_in_filter + (int) i + 1, no_default); \
		if (track_bad_kappa) { \
			if((*current_median == default_kappa) || (default_to_median && *num_bad_in_avg_re >= avg_array_size)) \
				new_avg = 0.0; \
			else \
				new_avg = 1.0; \
		} \
		src++; \
		/* Put data in the output buffer if there is room */ \
		if(dst_size + i >= src_size) { \
			*dst = new_avg; \
			dst++; \
		} \
	} \
	/* Update number of samples in running median and average */ \
	*samples_in_filter += src_size; \
	if(*samples_in_filter >= array_size + avg_array_size) \
		*samples_in_filter = array_size + avg_array_size - 1; \
	*total_samples += src_size; \
 \
	return GST_FLOW_OK; \
}


#define DEFINE_SMOOTH_COMPLEX_BUFFER(DTYPE) \
static GstFlowReturn smooth_complex_buffer_ ## DTYPE(const DTYPE complex *src, guint64 src_size, DTYPE complex *dst, guint64 dst_size, double *fifo_array_re, double *fifo_array_im, double *avg_array_re, double *avg_array_im, double default_kappa_re, double default_kappa_im, double *current_median_re, double *current_median_im, double maximum_offset_re, double maximum_offset_im, gint array_size, gint avg_array_size, int *index_re, int *index_im, int *avg_index_re, int *avg_index_im, int *num_bad_in_avg_re, int *num_bad_in_avg_im, gboolean gap, gboolean default_to_median, gboolean track_bad_kappa, int *samples_in_filter, gboolean no_default, guint64 *total_samples, gboolean reject_zeros, int *future_unused) { \
	guint64 i; \
	double new_element_re, new_element_im; \
	DTYPE complex new_avg; \
	for(i = 0; i < src_size; i++) { \
		double complex doublesrc = (double complex) *src; \
		if(gap || creal(doublesrc) > default_kappa_re + maximum_offset_re || creal(doublesrc) < default_kappa_re - maximum_offset_re || isnan(creal(doublesrc)) || isinf(creal(doublesrc)) || (reject_zeros && creal(doublesrc) == 0)) { \
			if(default_to_median) \
				new_element_re = *current_median_re; \
			else \
				new_element_re = default_kappa_re; \
 \
			if(*num_bad_in_avg_re <= avg_array_size) \
				(*num_bad_in_avg_re)++; \
		} else { \
			new_element_re = creal(doublesrc); \
			*num_bad_in_avg_re = 0; \
		} \
		if(gap || cimag(doublesrc) > default_kappa_im + maximum_offset_im || cimag(doublesrc) < default_kappa_im - maximum_offset_im || isnan(cimag(doublesrc)) || isinf(cimag(doublesrc)) || (reject_zeros && cimag(doublesrc) == 0)) { \
			if(default_to_median) \
				new_element_im = *current_median_im; \
			else \
				new_element_im = default_kappa_im; \
 \
			if(*num_bad_in_avg_im <= avg_array_size) \
				(*num_bad_in_avg_im)++; \
		} else { \
			new_element_im = cimag(doublesrc); \
			*num_bad_in_avg_im = 0; \
		} \
		/* Compute new median */ \
		get_new_median(new_element_re, fifo_array_re, current_median_re, array_size, index_re, index_im, FALSE, *samples_in_filter + (int) i + 1, no_default, *future_unused); \
		get_new_median(new_element_im, fifo_array_im, current_median_im, array_size, index_re, index_im, TRUE, *samples_in_filter + (int) i + 1, no_default, *future_unused); \
		if(*future_unused > 0) \
			(*future_unused)--; \
		/* Compute new average */ \
		new_avg = (DTYPE) get_average(*current_median_re, avg_array_re, avg_array_size, avg_index_re, avg_index_im, FALSE, *samples_in_filter + (int) i + 1, no_default) + I * (DTYPE) get_average(*current_median_im, avg_array_im, avg_array_size, avg_index_re, avg_index_im, TRUE, *samples_in_filter + (int) i + 1, no_default); \
		if(track_bad_kappa) { \
			if((*current_median_re == default_kappa_re || (default_to_median && *num_bad_in_avg_re >= avg_array_size)) && (*current_median_im == default_kappa_im || (default_to_median && *num_bad_in_avg_im >= avg_array_size))) \
				new_avg = 0.0; \
			else if(*current_median_im == default_kappa_im || (default_to_median && *num_bad_in_avg_im >= avg_array_size)) \
				new_avg = 1.0; \
			else if(*current_median_re == default_kappa_re || (default_to_median && *num_bad_in_avg_re >= avg_array_size)) \
				new_avg = I; \
			else \
				new_avg = 1.0 + I; \
		} \
		src++; \
		/* Put data in the output buffer if there is room */ \
		if(dst_size + i >= src_size) { \
			*dst = new_avg; \
			dst++; \
		} \
	} \
	/* Update number of samples in running median and average */ \
	*samples_in_filter += src_size; \
	if(*samples_in_filter >= array_size + avg_array_size) \
		*samples_in_filter = array_size + avg_array_size - 1; \
	*total_samples += src_size; \
 \
	return GST_FLOW_OK; \
}


DEFINE_SMOOTH_BUFFER(float);
DEFINE_SMOOTH_BUFFER(double);
DEFINE_SMOOTH_COMPLEX_BUFFER(float);
DEFINE_SMOOTH_COMPLEX_BUFFER(double);


/*
 * ============================================================================
 *
 *		     GstBaseTransform Method Overrides
 *
 * ============================================================================
 */


/*
 * get_unit_size()
 */


static gboolean get_unit_size(GstBaseTransform *trans, GstCaps *caps, gsize *size)
{
	GstAudioInfo info;
	gboolean success = gstlal_audio_info_from_caps(&info, caps);

	if(success) {
		*size = GST_AUDIO_INFO_BPF(&info);
	} else
		GST_WARNING_OBJECT(trans, "unable to parse caps %" GST_PTR_FORMAT, caps);

	return success;
}


/*
 * set_caps()
 */


static gboolean set_caps(GstBaseTransform *trans, GstCaps *incaps, GstCaps *outcaps)
{
	GSTLALSmoothKappas *element = GSTLAL_SMOOTHKAPPAS(trans);
	gboolean success = TRUE;
	gsize unit_size;
	gint rate_in, rate_out;

	/*
	 * parse the caps
	 */

	success &= get_unit_size(trans, incaps, &unit_size);
	GstStructure *str = gst_caps_get_structure(incaps, 0);
	const gchar *name = gst_structure_get_string(str, "format");
	success &= (name != NULL);
	success &= gst_structure_get_int(str, "rate", &rate_in);
	success &= gst_structure_get_int(gst_caps_get_structure(outcaps, 0), "rate", &rate_out);
	if(!success)
		GST_ERROR_OBJECT(element, "unable to parse caps.  input caps = %" GST_PTR_FORMAT " output caps = %" GST_PTR_FORMAT, incaps, outcaps);

	/* require the input and output rates to be equal */
	success &= (rate_in == rate_out);
	if(rate_in != rate_out)
		GST_ERROR_OBJECT(element, "output rate is not equal to input rate.  input caps = %" GST_PTR_FORMAT " output caps = %" GST_PTR_FORMAT, incaps, outcaps);

	/*
	 * record stream parameters
	 */

	if(success) {
		if(!strcmp(name, GST_AUDIO_NE(F32))) {
			element->data_type = GSTLAL_SMOOTHKAPPAS_F32;
			g_assert_cmpuint(unit_size, ==, 4);
		} else if(!strcmp(name, GST_AUDIO_NE(F64))) {
			element->data_type = GSTLAL_SMOOTHKAPPAS_F64;
			g_assert_cmpuint(unit_size, ==, 8);
		} else if(!strcmp(name, GST_AUDIO_NE(Z64))) {
			element->data_type = GSTLAL_SMOOTHKAPPAS_Z64;
			g_assert_cmpuint(unit_size, ==, 8);
		} else if(!strcmp(name, GST_AUDIO_NE(Z128))) {
			element->data_type = GSTLAL_SMOOTHKAPPAS_Z128;
			g_assert_cmpuint(unit_size, ==, 16);
		} else
			g_assert_not_reached();

		element->unit_size = unit_size;
		element->rate = rate_in;
	}

	return success;
}


/*
 * start()
 */


static gboolean start(GstBaseTransform *trans)
{
	GSTLALSmoothKappas *element = GSTLAL_SMOOTHKAPPAS(trans);

	element->current_median_re = element->default_kappa_re;
	element->current_median_im = element->default_kappa_im;

	element->fifo_array_re = g_malloc(sizeof(double) * element->array_size);
	element->fifo_array_im = g_malloc(sizeof(double) * element->array_size);
	element->avg_array_re = g_malloc(sizeof(double) * element->avg_array_size);
	element->avg_array_im = g_malloc(sizeof(double) * element->avg_array_size);

	int i;
	for(i = 0; i < element->array_size; i++) { 
		(element->fifo_array_re)[i] = element->default_kappa_re;
		(element->fifo_array_im)[i] = element->default_kappa_im;
	}
	for(i = 0; i < element->avg_array_size; i++) {
		(element->avg_array_re)[i] = element->default_kappa_re;
		(element->avg_array_im)[i] = element->default_kappa_im;
	}

	element->need_discont = TRUE;

	return TRUE;
}


/*
 * sink_event()
 */


static gboolean sink_event(GstBaseTransform *trans, GstEvent *event) {
	GSTLALSmoothKappas *element = GSTLAL_SMOOTHKAPPAS(trans);
	gboolean success = TRUE;
	GST_DEBUG_OBJECT(element, "Got %s event on sink pad", GST_EVENT_TYPE_NAME(event));

	int waste_samples = (int) (element->filter_latency * (element->array_size + element->avg_array_size - 2));
	waste_samples = waste_samples > 0 ? waste_samples : 0;
	if(GST_EVENT_TYPE(event) == GST_EVENT_EOS && waste_samples > 0) {
		/* Trick the function by passing it fake input data */
		void *fake = g_malloc(waste_samples * element->unit_size);
		void *data = g_malloc(waste_samples * element->unit_size);
		GstFlowReturn result;
		if(element->data_type == GSTLAL_SMOOTHKAPPAS_F32) {
			result = smooth_buffer_float((float *) fake, (guint64) waste_samples, (float *) data, (guint64) waste_samples, element->fifo_array_re, element->avg_array_re, element->default_kappa_re, &element->current_median_re, element->maximum_offset_re, element->array_size, element->avg_array_size, &element->index_re, &element->index_im, &element->avg_index_re, &element->avg_index_im, &element->num_bad_in_avg_re, TRUE, element->default_to_median, element->track_bad_kappa, &element->samples_in_filter, element->no_default, &element->total_samples, element->reject_zeros, &element->future_unused);
		} else if(element->data_type == GSTLAL_SMOOTHKAPPAS_F64) {
			result = smooth_buffer_double((double *) fake, (guint64) waste_samples, (double *) data, (guint64) waste_samples, element->fifo_array_re, element->avg_array_re, element->default_kappa_re, &element->current_median_re, element->maximum_offset_re, element->array_size, element->avg_array_size, &element->index_re, &element->index_im, &element->avg_index_re, &element->avg_index_im, &element->num_bad_in_avg_re, TRUE, element->default_to_median, element->track_bad_kappa, &element->samples_in_filter, element->no_default, &element->total_samples, element->reject_zeros, &element->future_unused);
		} else if(element->data_type == GSTLAL_SMOOTHKAPPAS_Z64) {
			result = smooth_complex_buffer_float((float complex *) fake, (guint64) waste_samples, (float complex *) data, (guint64) waste_samples, element->fifo_array_re, element->fifo_array_im, element->avg_array_re, element->avg_array_im, element->default_kappa_re, element->default_kappa_im, &element->current_median_re, &element->current_median_im, element->maximum_offset_re, element->maximum_offset_im, element->array_size, element->avg_array_size, &element->index_re, &element->index_im, &element->avg_index_re, &element->avg_index_im, &element->num_bad_in_avg_re, &element->num_bad_in_avg_im, TRUE, element->default_to_median, element->track_bad_kappa, &element->samples_in_filter, element->no_default, &element->total_samples, element->reject_zeros, &element->future_unused);
		} else if(element->data_type == GSTLAL_SMOOTHKAPPAS_Z128) {
			result = smooth_complex_buffer_double((double complex *) fake, (guint64) waste_samples, (double complex *) data, (guint64) waste_samples, element->fifo_array_re, element->fifo_array_im, element->avg_array_re, element->avg_array_im, element->default_kappa_re, element->default_kappa_im, &element->current_median_re, &element->current_median_im, element->maximum_offset_re, element->maximum_offset_im, element->array_size, element->avg_array_size, &element->index_re, &element->index_im, &element->avg_index_re, &element->avg_index_im, &element->num_bad_in_avg_re, &element->num_bad_in_avg_im, TRUE, element->default_to_median, element->track_bad_kappa, &element->samples_in_filter, element->no_default, &element->total_samples, element->reject_zeros, &element->future_unused);
		} else {
			result = GST_FLOW_ERROR;
			success = FALSE;
		}
		g_free(fake);
		fake = NULL;

		if(result == GST_FLOW_OK) {
			GstBuffer *buf;
			buf = gst_buffer_new_wrapped(data, (guint64) waste_samples * element->unit_size);

			set_metadata(element, buf, (guint64) waste_samples);

			/* push buffer downstream */
			GST_DEBUG_OBJECT(element, "pushing final buffer %" GST_BUFFER_BOUNDARIES_FORMAT, GST_BUFFER_BOUNDARIES_ARGS(buf));
			result = gst_pad_push(element->srcpad, buf);
		}
		if(G_UNLIKELY(result != GST_FLOW_OK)) {
			GST_WARNING_OBJECT(element, "push failed: %s", gst_flow_get_name(result));
			success = FALSE;
		}
	} else if (GST_EVENT_TYPE(event) == GST_EVENT_STREAM_START) {
		/* FIXME: This just makes a group_id to avoid FIXME's at stream start.  I don't think it implements group_id's in GStreamer's intended manner. */
		guint group_id;
		if(!gst_event_parse_group_id(event, &group_id)) {
			event = gst_event_make_writable(event);
			group_id = gst_util_group_id_next();
			gst_event_set_group_id(event, group_id);
		}
	}

	success &= GST_BASE_TRANSFORM_CLASS(gstlal_smoothkappas_parent_class)->sink_event(trans, event);

	return success;
}


/*
 * transform_size()
 */


static gboolean transform_size(GstBaseTransform *trans, GstPadDirection direction, GstCaps *caps, gsize size, GstCaps *othercaps, gsize *othersize)
{
	GSTLALSmoothKappas *element = GSTLAL_SMOOTHKAPPAS(trans);

	gsize unit_size;
	if(!get_unit_size(trans, caps, &unit_size)) {
		GST_DEBUG_OBJECT(element, "function 'get_unit_size' failed");
		return FALSE;
	}

	/* buffer size in bytes should be a multiple of unit_size in bytes */
	if(G_UNLIKELY(size % unit_size)) {
		GST_DEBUG_OBJECT(element, "buffer size %" G_GSIZE_FORMAT " is not a multiple of %" G_GSIZE_FORMAT, size, unit_size);
		return FALSE;
	}

	size /= unit_size;

	/* How many samples do we need to throw away based on the filter latency? */
	int waste_samples = (int) (element->filter_latency * (element->array_size + element->avg_array_size - 2));
	waste_samples = waste_samples > 0 ? waste_samples : 0;

	switch(direction) {
	case GST_PAD_SRC:
		/* We have the size of the output buffer, and we set the size of the input buffer. */
		/* Check if we need to clip the output buffer */
		if(element->total_samples >= (guint64) waste_samples)
			*othersize = size;
		else
			*othersize = size + waste_samples - element->total_samples;

		break;

	case GST_PAD_SINK:
		/* We have the size of the input buffer, and we set the size of the output buffer. */
		/* Check if we need to clip the output buffer */
		if(element->total_samples >= (guint64) waste_samples)
			*othersize = size;
		else if(size > (guint) (waste_samples - element->total_samples))
			*othersize = size - waste_samples + element->total_samples;
		else
			*othersize = 0;

		break;

	case GST_PAD_UNKNOWN:
		GST_ELEMENT_ERROR(trans, CORE, NEGOTIATION, (NULL), ("invalid direction GST_PAD_UNKNOWN"));
		return FALSE;
	}

	*othersize *= unit_size;

	return TRUE;
}


/*
 * transform()
 */


static GstFlowReturn transform(GstBaseTransform *trans, GstBuffer *inbuf, GstBuffer *outbuf)
{
	GSTLALSmoothKappas *element = GSTLAL_SMOOTHKAPPAS(trans);
	GstMapInfo inmap, outmap;
	GstFlowReturn result;
	int i, waste_samples;

	/* How many samples do we need to throw away based on the filter latency? */
	waste_samples = (int) (element->filter_latency * (element->array_size + element->avg_array_size - 2));

	/*
	 * Check for discontinuity
	 */

	if(G_UNLIKELY(GST_BUFFER_IS_DISCONT(inbuf) || GST_BUFFER_OFFSET(inbuf) != element->next_in_offset || !GST_CLOCK_TIME_IS_VALID(element->t0))) {
		element->offset0 = element->next_out_offset = GST_BUFFER_OFFSET(inbuf) + (waste_samples < 0 ? (guint64) (-waste_samples) : 0);
		element->t0 = GST_BUFFER_PTS(inbuf) + (waste_samples < 0 ? (guint64) (((-waste_samples) * GST_SECOND + 0.5) / element->rate) : 0);
		element->need_discont = TRUE;
	}
	element->next_in_offset = GST_BUFFER_OFFSET_END(inbuf);

	gboolean gap = GST_BUFFER_FLAG_IS_SET(inbuf, GST_BUFFER_FLAG_GAP);
	GST_INFO_OBJECT(element, "processing %s%s buffer %p spanning %" GST_BUFFER_BOUNDARIES_FORMAT, gap ? "gap" : "nongap", GST_BUFFER_FLAG_IS_SET(inbuf, GST_BUFFER_FLAG_DISCONT) ? "+discont" : "", inbuf, GST_BUFFER_BOUNDARIES_ARGS(inbuf));

	if(gap && element->clear_on_gap) {
		element->samples_in_filter = 0;
		element->index_re = 0;
		element->index_im = 0;
		element->current_median_re = element->default_kappa_re;
		element->current_median_im = element->default_kappa_im;

		for(i = 0; i < element->array_size; i++) {
			element->fifo_array_re[i] = element->default_kappa_re;
			element->fifo_array_im[i] = element->default_kappa_im;
		}
		for(i = 0; i < element->avg_array_size; i++) {
			element->avg_array_re[i] = element->default_kappa_re;
			element->avg_array_im[i] = element->default_kappa_im;
		}
	}
	/* Reset the running median and mean calculation if requested at this time */
	g_mutex_lock(&element->reset_mutex);
	if(element->num_reset_times > 0) {
		if(GST_BUFFER_PTS(inbuf) >= *element->reset_times) {
			if(element->min_array_size < element->samples_in_filter && element->min_array_size > 0)
				reset(element);
			element->num_reset_times--;
			for(i = 0; (guint) i < element->num_reset_times; i++)
				element->reset_times[i] = element->reset_times[i + 1];
			element->reset_times = g_realloc(element->reset_times, element->num_reset_times * sizeof(guint64));
		}
	}
	g_mutex_unlock(&element->reset_mutex);

	gst_buffer_map(outbuf, &outmap, GST_MAP_WRITE);
	gst_buffer_map(inbuf, &inmap, GST_MAP_READ);

	/* sanity checks */
	g_assert_cmpuint(inmap.size % element->unit_size, ==, 0);
	g_assert_cmpuint(outmap.size % element->unit_size, ==, 0);

	/* Process data in buffer */
	if(element->data_type == GSTLAL_SMOOTHKAPPAS_F32) {
		result = smooth_buffer_float((const float *) inmap.data, inmap.size / element->unit_size, (float *) outmap.data, outmap.size / element->unit_size, element->fifo_array_re, element->avg_array_re, element->default_kappa_re, &element->current_median_re, element->maximum_offset_re, element->array_size, element->avg_array_size, &element->index_re, &element->index_im, &element->avg_index_re, &element->avg_index_im, &element->num_bad_in_avg_re, gap, element->default_to_median, element->track_bad_kappa, &element->samples_in_filter, element->no_default, &element->total_samples, element->reject_zeros, &element->future_unused);
	} else if(element->data_type == GSTLAL_SMOOTHKAPPAS_F64) {
		result = smooth_buffer_double((const double *) inmap.data, inmap.size / element->unit_size, (double *) outmap.data, outmap.size / element->unit_size, element->fifo_array_re, element->avg_array_re, element->default_kappa_re, &element->current_median_re, element->maximum_offset_re, element->array_size, element->avg_array_size, &element->index_re, &element->index_im, &element->avg_index_re, &element->avg_index_im, &element->num_bad_in_avg_re, gap, element->default_to_median, element->track_bad_kappa, &element->samples_in_filter, element->no_default, &element->total_samples, element->reject_zeros, &element->future_unused);
	} else if(element->data_type == GSTLAL_SMOOTHKAPPAS_Z64) {
		result = smooth_complex_buffer_float((const float complex *) inmap.data, inmap.size / element->unit_size, (float complex *) outmap.data, outmap.size / element->unit_size, element->fifo_array_re, element->fifo_array_im, element->avg_array_re, element->avg_array_im, element->default_kappa_re, element->default_kappa_im, &element->current_median_re, &element->current_median_im, element->maximum_offset_re, element->maximum_offset_im, element->array_size, element->avg_array_size, &element->index_re, &element->index_im, &element->avg_index_re, &element->avg_index_im, &element->num_bad_in_avg_re, &element->num_bad_in_avg_im, gap, element->default_to_median, element->track_bad_kappa, &element->samples_in_filter, element->no_default, &element->total_samples, element->reject_zeros, &element->future_unused);
	} else if(element->data_type == GSTLAL_SMOOTHKAPPAS_Z128) {
		result = smooth_complex_buffer_double((const double complex *) inmap.data, inmap.size / element->unit_size, (double complex *) outmap.data, outmap.size / element->unit_size, element->fifo_array_re, element->fifo_array_im, element->avg_array_re, element->avg_array_im, element->default_kappa_re, element->default_kappa_im, &element->current_median_re, &element->current_median_im, element->maximum_offset_re, element->maximum_offset_im, element->array_size, element->avg_array_size, &element->index_re, &element->index_im, &element->avg_index_re, &element->avg_index_im, &element->num_bad_in_avg_re, &element->num_bad_in_avg_im, gap, element->default_to_median, element->track_bad_kappa, &element->samples_in_filter, element->no_default, &element->total_samples, element->reject_zeros, &element->future_unused);
	} else
		g_assert_not_reached();

	set_metadata(element, outbuf, outmap.size / element->unit_size);

	gst_buffer_unmap(inbuf, &inmap);

	gst_buffer_unmap(outbuf, &outmap);

	/*
	 * done
	 */

	return result;
}


/*
 * ============================================================================
 *
 *			  GObject Method Overrides
 *
 * ============================================================================
 */


/*
 * properties
 */


enum property {
	ARG_ARRAY_SIZE = 1,
	ARG_MIN_ARRAY_SIZE,
	ARG_AVG_ARRAY_SIZE,
	ARG_DEFAULT_KAPPA_RE,
	ARG_DEFAULT_KAPPA_IM,
	ARG_MAXIMUM_OFFSET_RE,
	ARG_MAXIMUM_OFFSET_IM,
	ARG_DEFAULT_TO_MEDIAN,
	ARG_TRACK_BAD_KAPPA,
	ARG_FILTER_LATENCY,
	ARG_NO_DEFAULT,
	ARG_CLEAR_ON_GAP,
	ARG_RESET_TIME,
	ARG_REJECT_ZEROS
};


static void set_property(GObject *object, enum property prop_id, const GValue *value, GParamSpec *pspec)
{
	GSTLALSmoothKappas *element = GSTLAL_SMOOTHKAPPAS(object);

	GST_OBJECT_LOCK(element);

	switch (prop_id) {
	case ARG_ARRAY_SIZE:
		element->array_size = g_value_get_int(value);
		break;
	case ARG_MIN_ARRAY_SIZE:
		element->min_array_size = g_value_get_int(value);
		break;
	case ARG_AVG_ARRAY_SIZE:
		element->avg_array_size = g_value_get_int(value);
		break;
	case ARG_DEFAULT_KAPPA_RE:
		element->default_kappa_re = g_value_get_double(value);
		break;
	case ARG_DEFAULT_KAPPA_IM:
		element->default_kappa_im = g_value_get_double(value);
		break;
	case ARG_MAXIMUM_OFFSET_RE:
		element->maximum_offset_re = g_value_get_double(value);
		break;
	case ARG_MAXIMUM_OFFSET_IM:
		element->maximum_offset_im = g_value_get_double(value);
		break;
	case ARG_DEFAULT_TO_MEDIAN:
		element->default_to_median = g_value_get_boolean(value);
		break;
	case ARG_TRACK_BAD_KAPPA:
		element->track_bad_kappa = g_value_get_boolean(value);
		break;
	case ARG_FILTER_LATENCY:
		element->filter_latency = g_value_get_double(value);
		break;
	case ARG_NO_DEFAULT:
		element->no_default = g_value_get_boolean(value);
		break;
	case ARG_CLEAR_ON_GAP:
		element->clear_on_gap = g_value_get_boolean(value);
		break;
	case ARG_RESET_TIME: ;
		guint64 old_reset_time, new_reset_time = g_value_get_uint64(value);
		guint i = 0;
		/* Only store nonzero reset times that we don't already have */
		gboolean store_new_reset_time = new_reset_time > 0;
		if(store_new_reset_time) {
			for(i = 0; i < element->num_reset_times; i++)
				store_new_reset_time = store_new_reset_time && new_reset_time != element->reset_times[i];
		}
		if(store_new_reset_time) {
			g_mutex_lock(&element->reset_mutex);
			element->num_reset_times++;
			element->reset_times = g_realloc(element->reset_times, element->num_reset_times * sizeof(guint64));
			/* Store the reset times in increasing order */
			while(element->reset_times[i] < new_reset_time && i < element->num_reset_times - 1)
				i++;
			old_reset_time = element->reset_times[i];
			element->reset_times[i] = new_reset_time;
			/* Shift values in the array if we had to insert the new reset time in the middle */
			while(i < element->num_reset_times - 1) {
				i++;
				new_reset_time = element->reset_times[i];
				element->reset_times[i] = old_reset_time;
				old_reset_time = new_reset_time;
			}
			g_mutex_unlock(&element->reset_mutex);
			if(element->min_array_size == 0)
				GST_WARNING_OBJECT(element, "received a reset time, but min_array_size is not set, so reset will not occur");
		}
		break;
	case ARG_REJECT_ZEROS:
		element->reject_zeros = g_value_get_boolean(value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		break;
	}

	GST_OBJECT_UNLOCK(element);
}


static void get_property(GObject *object, enum property prop_id, GValue *value, GParamSpec *pspec)
{
	GSTLALSmoothKappas *element = GSTLAL_SMOOTHKAPPAS(object);

	GST_OBJECT_LOCK(element);

	switch (prop_id) {
	case ARG_ARRAY_SIZE:
		g_value_set_int(value, element->array_size);
		break;
	case ARG_MIN_ARRAY_SIZE:
		g_value_set_int(value, element->min_array_size);
		break;
	case ARG_AVG_ARRAY_SIZE:
		g_value_set_int(value, element->avg_array_size);
		break;
	case ARG_DEFAULT_KAPPA_RE:
		g_value_set_double(value, element->default_kappa_re);
		break;
	case ARG_DEFAULT_KAPPA_IM:
		g_value_set_double(value, element->default_kappa_im);
		break;
	case ARG_MAXIMUM_OFFSET_RE:
		g_value_set_double(value, element->maximum_offset_re);
		break;
	case ARG_MAXIMUM_OFFSET_IM:
		g_value_set_double(value, element->maximum_offset_im);
		break;
	case ARG_DEFAULT_TO_MEDIAN:
		g_value_set_boolean(value, element->default_to_median);
		break;
	case ARG_TRACK_BAD_KAPPA:
		g_value_set_boolean(value, element->track_bad_kappa);
		break;
	case ARG_FILTER_LATENCY:
		g_value_set_double(value, element->filter_latency);
		break;
	case ARG_NO_DEFAULT:
		g_value_set_boolean(value, element->no_default);
		break;
	case ARG_CLEAR_ON_GAP:
		g_value_set_boolean(value, element->clear_on_gap);
		break;
	case ARG_RESET_TIME:
		if(element->num_reset_times > 0)
			g_value_set_uint64(value, element->reset_times[0]);
		else
			g_value_set_uint64(value, 0);
		break;
	case ARG_REJECT_ZEROS:
		g_value_set_boolean(value, element->reject_zeros);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
		break;
	}

	GST_OBJECT_UNLOCK(element);
}


/*
 * finalize()
 */


static void finalize(GObject *object)
{
	GSTLALSmoothKappas *element = GSTLAL_SMOOTHKAPPAS(object);
	gst_object_unref(element->srcpad);
	element->srcpad = NULL;
	g_free(element->fifo_array_re);
	element->fifo_array_re = NULL;
	g_free(element->fifo_array_im);
	element->fifo_array_im = NULL;
	g_free(element->avg_array_re);
	element->avg_array_re = NULL;
	g_free(element->avg_array_im);
	element->avg_array_im = NULL;
	if(element->reset_times) {
		g_free(element->reset_times);
		element->reset_times = NULL;
		element->num_reset_times = 0;
	}
	g_mutex_clear(&element->reset_mutex);
	G_OBJECT_CLASS(gstlal_smoothkappas_parent_class)->finalize(object);
}


/*
 * class_init()
 */


static void gstlal_smoothkappas_class_init(GSTLALSmoothKappasClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	GstElementClass *element_class = GST_ELEMENT_CLASS(klass);
	GstBaseTransformClass *transform_class = GST_BASE_TRANSFORM_CLASS(klass);

	gst_element_class_set_details_simple(
		element_class,
		"Smooth Calibration Factors",
		"Filter/Audio",
		"Smooths the calibration factors with a running median.",
		"Madeline Wade <madeline.wade@ligo.org>, Aaron Viets <aaron.viets@ligo.org>"
	);

	gobject_class->set_property = GST_DEBUG_FUNCPTR(set_property);
	gobject_class->get_property = GST_DEBUG_FUNCPTR(get_property);
	gobject_class->finalize = GST_DEBUG_FUNCPTR(finalize);

	transform_class->get_unit_size = GST_DEBUG_FUNCPTR(get_unit_size);
	transform_class->set_caps = GST_DEBUG_FUNCPTR(set_caps);
	transform_class->start = GST_DEBUG_FUNCPTR(start);
	transform_class->sink_event = GST_DEBUG_FUNCPTR(sink_event);
	transform_class->transform_size = GST_DEBUG_FUNCPTR(transform_size);
	transform_class->transform = GST_DEBUG_FUNCPTR(transform);

	gst_element_class_add_pad_template(element_class, gst_static_pad_template_get(&src_factory));
	gst_element_class_add_pad_template(element_class, gst_static_pad_template_get(&sink_factory));

	g_object_class_install_property(
		gobject_class,
		ARG_ARRAY_SIZE,
		g_param_spec_int(
			"array-size",
			"Median array size",
			"Size of the array of values from which the median is calculated.",
			0, G_MAXINT, 2049,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_MIN_ARRAY_SIZE,
		g_param_spec_int(
			"min-array-size",
			"Minimum Median Array Size",
			"Number of samples to be retained in median array when element is reset",
			0, G_MAXINT, 0,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_AVG_ARRAY_SIZE,
		g_param_spec_int(
			"avg-array-size",
			"Average array size",
			"Size of the array of values from which the average is calculated\n\t\t\t"
			"from the median values. By default, no average is taken.",
			0, G_MAXINT, 1,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_DEFAULT_KAPPA_RE,
		g_param_spec_double(
			"default-kappa-re",
			"Default real part of kappa value",
			"Default real part of kappa value to be used if there is a gap in the\n\t\t\t"
			"incoming buffer, or if no input values pass kappa-offset criteria.\n\t\t\t"
			"All elements of the real fifo array are initialized to this value.",
			-G_MAXDOUBLE, G_MAXDOUBLE, 1.0,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_DEFAULT_KAPPA_IM,
		g_param_spec_double(
			"default-kappa-im",
			"Default imaginary part of kappa value",
			"Default imaginary part of kappa value to be used if there is a gap in the\n\t\t\t"
			"incoming buffer, or if no input values pass kappa-offset criteria. All\n\t\t\t"
			"elements of the imaginary fifo array are initialized to this value.",
			-G_MAXDOUBLE, G_MAXDOUBLE, 0.0,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_MAXIMUM_OFFSET_RE,
		g_param_spec_double(
			"maximum-offset-re",
			"Maximum acceptable real kappa offset",
			"Maximum acceptable offset of unsmoothed real kappa from default-kappa-re\n\t\t\t"
			"to be entered into real array from which median is calculated.",
			0, G_MAXDOUBLE, G_MAXFLOAT / 2,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_MAXIMUM_OFFSET_IM,
		g_param_spec_double(
			"maximum-offset-im",
			"Maximum acceptable imaginary kappa offset",
			"Maximum acceptable offset of unsmoothed imaginary kappa from default-kappa-im\n\t\t\t"
			"to be entered into imaginary-part array from which median is calculated.",
			0, G_MAXDOUBLE, G_MAXFLOAT / 2,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_DEFAULT_TO_MEDIAN,
		g_param_spec_boolean(
			"default-to-median",
			"Default to median",
			"If set to false (default), gaps (or times where input values do not pass\n\t\t\t"
			"kappa-offset criteria) are filled in by entering default-kappa into the\n\t\t\t"
			"fifo array. If set to true, gaps are filled in by entering the current\n\t\t\t"
			"median value into the fifo array.",
			FALSE,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_TRACK_BAD_KAPPA,
		g_param_spec_boolean(
			"track-bad-kappa",
			"Track input bad kappas",
			"If set to false (default), gaps (or times where input values do not pass\n\t\t\t"
			"kappa-offset criteria) are filled in by entering default-kappa into the fifo\n\t\t\t"
			"array and non-gaps use the input buffer value. If set to true, gaps are\n\t\t\t"
			"filled in by entering 0 into the fifo array and non-gaps are filled by\n\t\t\t"
			"entering 1's into the fifo array.",
			FALSE,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_FILTER_LATENCY,
		g_param_spec_double(
			"filter-latency",
			"Filter Latency",
			"The latency associated with the smoothing process, as a fraction of the\n\t\t\t"
			"total length of the running median + average. If 0, there is no latency.\n\t\t\t"
			"If 1, the latency is the length of the running median + average.",
			-G_MAXDOUBLE, G_MAXDOUBLE, 0.0,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_NO_DEFAULT,
		g_param_spec_boolean(
			"no-default",
			"No Default Kappa",
			"If set to true, default values will never be used. Instead, only actual\n\t\t\t"
			"inputs are used in the median and average arrays, so that the true median or\n\t\t\t"
			"average is reported even if the array is not full yet.",
			FALSE,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_CLEAR_ON_GAP,
		g_param_spec_boolean(
			"clear-on-gap",
			"Clear Memory on Gap",
			"If set to true, all values present in the median and average arrays will be\n\t\t\t"
			"deleted when the input is a gap, and the calculation will start over when\n\t\t\t"
			"the next nongap data arrives.",
			FALSE,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_RESET_TIME,
		g_param_spec_uint64(
			"reset-time",
			"Reset Time",
			"Next presentation timestamp (in ns) at which values present in the median\n\t\t\t"
			"and average arrays will be deleted (min array sizes are kept), and the\n\t\t\t"
			"calculation will start over.  Default is to not reset.",
			0, G_MAXUINT64, 0,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT | GST_PARAM_CONTROLLABLE
		)
	);
	g_object_class_install_property(
		gobject_class,
		ARG_REJECT_ZEROS,
		g_param_spec_boolean(
			"reject-zeros",
			"Reject Zeros",
			"If true, zeros are replaced with the default value or current median.",
			TRUE,
			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT
		)
	);
}


/*
 * init()
 */


static void gstlal_smoothkappas_init(GSTLALSmoothKappas *element) {

	/* retrieve (and ref) src pad */
	GstPad *pad = gst_element_get_static_pad(GST_ELEMENT(element), "src");
	GST_PAD_SET_PROXY_CAPS(pad);
	GST_PAD_SET_PROXY_ALLOCATION(pad);
	GST_PAD_SET_PROXY_SCHEDULING(pad);
	element->srcpad = pad;

	element->unit_size = 0;
	element->rate = 0;
	element->array_size = 0;
	element->min_array_size = 0;
	element->avg_array_size = 0;
	element->fifo_array_re = NULL;
	element->fifo_array_im = NULL;
	element->avg_array_re = NULL;
	element->avg_array_im = NULL;
	element->index_re = 0;
	element->index_im = 0;
	element->avg_index_re = 0;
	element->avg_index_im = 0;
	element->num_bad_in_avg_re = G_MAXINT;
	element->num_bad_in_avg_im = G_MAXINT;
	element->samples_in_filter = 0;
	element->total_samples = 0;
	element->future_unused = 0;
	element->reset_times = NULL;
	element->num_reset_times = 0;

	g_mutex_init(&element->reset_mutex);

	gst_base_transform_set_qos_enabled(GST_BASE_TRANSFORM(element), TRUE);
	gst_base_transform_set_gap_aware(GST_BASE_TRANSFORM(element), TRUE);
}
